module Data.GraphRewriting.TermGraph        
where 
  
import Data.GraphRewriting.ExtendedEarl
import Data.Rewriting.Term.Type
import Data.List as List (foldr, foldl', map, maximum, length, partition, nub, concatMap, (\\))

-- | Represents a termgraph based on the graph being of type Earl, 
--   this enables a termgraph to instanciate ExtendedEarl, as it extends
--   an Earl only insofar, as it stores the root node. 
data TermGraph = TG { root  :: Node,  -- ^ the root of the TermGraph 
                      graph :: Earl   -- ^ the graph holding the information
                    } 

-- | Position represent a position within a TermGraph.
type Position = [Int]

instance ExtendedEarl TermGraph where
  toEarl = graph 
  insertNode (TG rt gr) ln  =  TG rt (einsertNode gr ln) 
  deleteNode (TG rt gr) n   =  TG rt (edeleteNode gr n)
  insertEdge (TG rt gr) le  =  TG rt (einsertEdge gr le)
  deleteEdge (TG rt gr) e   =  TG rt (edeleteEdge gr e)
  deleteLEdge (TG rt gr) le = TG rt (edeleteLEdge gr le)


-- | subTG (TG,n): TG|^n, i.e. the termgraph starting from Node n, hence it
--   deletes inaccesible nodes, also when applied to root. Note that: TG|^(root
--   tg) = TG, TG|^n = (root = n, graph = graph starting from n)
subTG :: (TermGraph, Node) -> TermGraph 
subTG (tg,newrt) = setroot (deleteNodes tg unreachable) newrt
  where unreachable = (nodes tg) List.\\ (reachabl tg newrt)
        setroot (TG _ gr) nrt  = (TG nrt gr)

-- alternative implementation, which does not delete inaccesible nodes, but reconstructs a graph from the accesible ones
--subTG (tg,newrt) = insertEdges nodestg edg
  --where (nodestg, edg) = foldc f tg (context tg newrt) ((TG newrt empty),[]) 
    --    f (_, n, ln, sucn) (tgi,edi) = (insertNode tgi (n,ln), ([(n,si,el) | (el, si) <- sucn])++edi)

(|^) :: TermGraph -> Node -> TermGraph
(|^) = curry subTG
  
-- | toTerm S: calculates to termgraph S the corresponding term
toTerm :: TermGraph -> Term Symbol Variable
toTerm tg = toTerm' (root tg)
   where toTerm' n = case label tg n of
           GVar x -> Var x
           GFun f -> Fun f [toTerm' s | s <- succs tg n]
                                

-- calculates a term graph or more accurately - a term tree - as no variable nodes are shared, depth first call.
toTermGraph :: Term Symbol Variable -> TermGraph
toTermGraph t = toTG (inittg t) 1 -- initialize succNo with 1  (succNo ^= how manyth successor is this)
     where toTG (tg,   _,     [])  _      = tg
           toTG (tg, anc, (ti:ts)) succNo = toTG (tgi,anc,ts) (succNo+1) 
             where tgi = case ti of 
                              Fun f tis -> toTG ((insert tg (fn,GFun f) anc succNo),fn,tis) 1 -- start with succNo == 1 again 
                              Var x     -> insert tg (fn,GVar x) anc succNo
                   fn = freshNode tg   
                   insert tg' labN@(n,_) anc' succNo' = insertEdge (insertNode tg' labN) (anc',n,succNo')
   
   
-- | toMinTermGraph t: calculates the minimally sharing termgraph, i.e. only
--   variable nodes are shared, corresponding to the given term t.
toMinTermGraph :: Term Symbol Variable -> TermGraph
toMinTermGraph t = shareVariableNodes (toTermGraph t)

-- a new termgraph where the root label is indicated by given term, returns (termgraph, root nodenr, succTerms)
inittg :: Term Symbol Variable -> (TermGraph,Node,[Term Symbol Variable])
inittg t  = case t of
                 Var x    -> (insertNode (TG fn empty) (fn,(GVar x)),fn,[])
                 Fun f ts -> (insertNode (TG fn empty) (fn,(GFun f)),fn,ts)
            where fn = freshNode empty

-- share variable nodes to produce a minimally shared termgraph
shareVariableNodes :: TermGraph -> TermGraph
shareVariableNodes tg = sharermin tg [n | n <- nodes tg, isVarNode tg n]
  
  
-------------------------------------------------------------------------------------------------------------------
-- * positioning, inspection
-------------------------------------------------------------------------------------------------------------------

-- | nodeATpos tg pos: returns the node corresponding to the Position pos in TermGraph tg
nodeATpos :: TermGraph -> Position -> Node 
nodeATpos tg pos = nodeATpos' (root tg) pos 
  where nodeATpos'  n []       = n
        nodeATpos'  n (p:pos') = nodeATpos' (ithsucc tg n p) pos'  

-- | possetOFnode tg n: gives the set of positions the Node in TermGraph tg is referred by.
possetOFnode :: TermGraph -> Node -> [Position]
possetOFnode tg n | n == root tg = [[]]
                  | otherwise    = List.foldr (\ (a,_,l) pos -> (List.map (++[l]) (possetOFnode tg a)) ++ pos) [] (inn tg n)

-- | height tg n: returns the length of the longest path in TermGraph tg from the root to Node n
height :: TermGraph -> Node -> Int
height tg n = List.maximum (List.map List.length (possetOFnode tg n))


-------------------------------------------------------------------------------------------------------------------
-- * folding and unfolding
-------------------------------------------------------------------------------------------------------------------

-- | unfold tg pos: unfolds (unshares) the TermGraph tg above Position pos,
--   hence afterwards the TermGraph will not contain any shared nodes above
--   the node corresponding to pos. This tackles the problem of accidentaly rewriting
--   two redices in a TermGraph simoultanesouly.
unfold :: TermGraph -> Position -> TermGraph
unfold tg []  = tg
unfold tg pos = unfold' tg (root tg) pos 
  where unfold' uftg  _   []  = uftg 
        unfold' uftg n (p:ps) = unfold' uftg' si' ps
          where si  = ithsucc uftg n p
                (uftg', si') | isDirectlyShared uftg si = unfoldNode uftg n si p
                             | otherwise                = (uftg,si)


-- | unfoldNode tg anc suci i: unfolds Node suci in TermGraph tg (by copying it)  with and 
--   redirects the edge from anc with edgelabel i (i.e. the ith egde) to the new Node
unfoldNode :: TermGraph -> Node -> Node -> EdgeLabel -> (TermGraph,Node)
unfoldNode tg anc suci i = (deleteLEdge (insertEdge tg' (anc,copiedsuci,i)) (anc,suci,i), copiedsuci) 
  where (tg',copiedsuci) = copy tg suci


-- sharermin tg rmin: in tg all nodes in rmin will be folded.
sharermin :: TermGraph -> [Node] -> TermGraph
sharermin tg [] = tg
sharermin tg (p:rmin) = sharermin (List.foldr (\q tgi -> foldNode tgi q p) tg homop) nothomop
  where (homop, nothomop) = partition (homomorph tg p) rmin 
        
-- | foldfromroot tg: folds the given termgraph tg from the root, i.e., computes a maximally sharing termgraph
foldfromroot :: TermGraph ->  TermGraph
foldfromroot tg = foldtg' tg (leafnodes tg)
  where foldtg' tgi [] = tgi
        foldtg' tgi ns = foldtg' ftg repancs
          where (ftg,reps) = foldequivclasses tgi (equivclasses tgi ns)
                repancs = List.nub (List.concatMap (ancs ftg) reps)

-- given a list of equivalence classes, folds all nodes in each equivalence class to one representative
-- returns the folded termgraph and the list of kept representatives
foldequivclasses :: TermGraph -> [[Node]] -> (TermGraph,[Node])
foldequivclasses tg nss = foldequivc tg nss [] 
  where foldequivc tg' [] reps = (tg',reps)
        foldequivc tg' ((n:ns):mss) reps = foldequivc (List.foldl' (\tgi q -> foldNode tgi q n) tg' ns) mss (n:reps)
        foldequivc _   ([]:_) _ = error "TermGraph.foldequivclasses: there are no empty equivalence classes of nodes."

-- partition a list of nodes into equivalance classes, using equivalence induced by homomorph
equivclasses :: TermGraph -> [Node] -> [[Node]]
equivclasses tg ms = equivc ms []
  where equivc [] ec = ec 
        equivc (n:ns) ec = equivc notEquivN ((n:equivN):ec)
          where (equivN,notEquivN) = List.partition (homomorph tg n) ns
 

-- | foldtg tg pos: folds (shares) all nodes in TermGraph tg which are below Position pos
--   hence afterwards the TermGraph will contain only shared nodes below the node corresponding
--   to the pos. This tackles the problem of not finding a Morphism
foldtgatpos :: TermGraph -> Position -> TermGraph
foldtgatpos tg pos = case (foldable (tg |^ nodeATpos tg pos)) of
   []         -> tg
   foldablens -> foldtgatpos (foldtg' tg foldablens) pos
  where foldtg' tg' []            =  tg'
        foldtg' tg' ((n:ns):rest) = foldtg' (List.foldl' (\tgi q -> foldNode tgi q n) tg' ns) rest
        foldtg' _   ([]:_) = error "TermGraph.foldtgatpos: there are no empty equivalence classes of nodes (foldable)."

-- finds all foldable nodes in a termgraph 
foldable :: TermGraph -> [[Node]]
foldable tg = fldable (nodes tg)  []
  where fldable []     grouped = grouped
        fldable (n:ns) grouped = fldable notHomoN grouped'
          where (homoN, notHomoN) = List.partition (homomorph tg n) ns
                grouped' | homoN == [] = grouped
                         | otherwise   = (n:homoN):grouped

-- | foldNode tg deln n: folds (shares) Node deln and n in TermGraph tg where it keeps
--   the Node n in the TermGraph.
foldNode :: TermGraph -> Node -> Node -> TermGraph
foldNode tg deln n = deleteNode (redirect tg deln n) deln
                 
-- checks whether the given position is in the given termgraph
validpos :: TermGraph -> Position -> Bool        
validpos tg pos = valpos (root tg) pos 
  where valpos _ [] = True
        valpos n (p:ps) | (List.length (out tg n) >= p) = True && valpos (ithsucc tg n p) ps
                        | otherwise = False
