{-# LANGUAGE TypeSynonymInstances, OverlappingInstances #-}

module Rewritingi (
   load_sys
 , load_term 
 , load_termgraph  
 , load_term_and_graph  
 , termgraph
 , term  
 , trs  
 , grs 
 , adv_gr_step  
 , adv_gr_steps  
 , adv_gr_nf  
 , adv_sharing   
 , adv_unsharing  
 , adv_t_step
 , adv_t_steps  
 , adv_t_nf  
 , gr_step  
 , gr_steps  
 , gr_nf  
 , sharing 
 , unsharing  
 , t_step 
 , t_steps 
 , t_nf
 , show_trs
 , help  
)

where 

import Data.IORef
import System.IO.Unsafe
import Text.PrettyPrint.ANSI.Leijen as Doc
  
-- graphrewriting
import Main (TermRewriteSystem, grsfromfile, tgfromstring)
import Data.GraphRewriting.GraphRewriteRule
import Data.GraphRewriting.TermGraph 
import qualified Data.GraphRewriting.Grewrite as Grewrite (informativeoutermostrewrite, outermostrewrite)
import Data.GraphRewriting.Printtt (printtg,printpos,printgrwrule)

-- termrewriting
import Data.Rewriting.Rule hiding (lhs, rhs)
import Data.Rewriting.Term (prettyTerm)
import qualified Data.Rewriting.Rules as TermRewriting (outerRewrite, result)

type UTerm = Term String String

instance Show TermGraph where 
  show tg = show $ printtg tg

-- For this to work, the term-rewriting library has to be installed with OverlappingInstances
-- If this is done, the below instance allows nicer term output.
-- instance Show UTerm where
--  show t = show (prettyTerm text text t)

-- | represents the current state of the programm
data STATE = STATE { ssys :: Maybe (TermRewriteSystem,GraphRewriteSystem) -- ^ the loaded rewrite system, init value: Nothing, corresponding (termrewritesystem,graphrewritesystem)
                   , sut  :: Maybe UTerm                                  -- ^ term set by user to be remembered
                   , sutg :: Maybe TermGraph                              -- ^ termgraph set by user to be remembered
                   } 
                              

-------------------------------------------------------------------------------
-- STATEHANDLING
  
stateRef :: IORef STATE
stateRef = unsafePerformIO $ newIORef (STATE Nothing Nothing Nothing)
{-# NOINLINE stateRef #-}

getState :: IO STATE
getState = readIORef stateRef
------------------------------------------------------------------------------


----------------------------------------------------------------
-- * load stored intitials
----------------------------------------------------------------
                              
-- | load fp: loads a termrewriting system from the given file path, computes the corresponding graphrewrite system 
--   and updates state with this values, i.e. leaves term and termgraph unchanged 
load_sys :: FilePath -> IO ()
load_sys fp = do
  (fgrs,ftrs) <- grsfromfile fp
  STATE _ t tg <- getState
  writeIORef stateRef (STATE (Just (ftrs,fgrs)) t tg)

-- | load_term: sets the term in the state, leaves other values unchanged.
load_term :: String -> IO UTerm
load_term s = do 
  (_, t) <- tgfromstring s
  STATE rwsys _ tg <- getState
  writeIORef stateRef (STATE rwsys (Just t) tg)
  return t

-- | load_termgraph: sets the termgraph in the state, leaves other values unchanged.
load_termgraph :: String -> IO TermGraph
load_termgraph s = do 
  (tg, _) <- tgfromstring s
  STATE rwsys t _ <- getState
  writeIORef stateRef (STATE rwsys t (Just tg))
  return tg

-- | load_term_and_graph: sets the term and the corresponding termgraph in the state  parallely, does not change the rewrite system
load_term_and_graph :: String -> IO (TermGraph, UTerm)
load_term_and_graph s = do 
  (tg, t) <- tgfromstring s
  STATE rwsys _ _ <- getState
  writeIORef stateRef (STATE rwsys (Just t) (Just tg))
  return (tg, t)

----------------------------------------------------------------
-- * return in STATE stored values
----------------------------------------------------------------
                              
-- | termgraph: returns stored termgraph
termgraph :: IO TermGraph
termgraph = do 
  STATE _ _ tg <- getState
  case tg of
    Just utg -> return utg
    Nothing  -> error "No termgraph loaded."
            
-- | term: returns the stored term
term :: IO UTerm
term = do 
  STATE _ t _ <- getState
  case t of
    Just ut -> return ut
    Nothing -> error "No term loaded."

-- | trs: returns the stored term rewrite system
trs :: IO TermRewriteSystem
trs = do 
  STATE rws _ _ <- getState
  case rws of
    Just (utrs,_) -> return utrs
    Nothing       -> error "No rewrite system loaded."

-- | grs: returns the stored graph rewrite system
grs :: IO GraphRewriteSystem
grs = do 
  STATE rws _ _ <- getState
  case rws of
    Just (_,ugrs) -> return ugrs
    Nothing       -> error "No rewrite system loaded."

-- | print the stored trs nicely
show_trs :: IO ()    
show_trs = do
  rules <- trs
  putDoc $ foldl (\d r -> d <> (prettyRule (text "->") text text r) <> line) Doc.empty rules
  
----------------------------------------------------------------
-- * advanced graph rewrite functions
----------------------------------------------------------------

-- prints additional information about the rewrite step
information :: Maybe (GraphRewriteRule, Position, TermGraph, TermGraph) -> IO ()
information Nothing = putStrLn "Termgraph is in normal form."
information (Just (rule, pos, sharedtg, _)) = putDoc (text "The shared termgraph: " <$> 
                                                      printtg sharedtg <$>
                                                      text "rewrites with" <$> printgrwrule rule <$>
                                                      text "at position"   <+>  printpos pos <+>
                                                      text "to"            <> line
                                                      )


-- | adv_gr_step ugrs utg: rewrites the termgraph utg wrt. graph rewrite system ugrs
adv_gr_step :: IO GraphRewriteSystem -> IO TermGraph -> IO TermGraph
adv_gr_step ugrs utg = do
  grw <- ugrs
  tg <- utg       
  let result = Grewrite.informativeoutermostrewrite grw tg  in
    case result of
      Just (_, _, _, tg') ->  putStrLn (show tg) >> information result >> return tg'
      Nothing -> information result >> return tg
      
-- | adv_gr_steps ugrs utg n: perfoms n graphrewrite steps with given termgraph utg wrt. graphrewritesystem ugrs, halts before if termgraph reached normalform   
adv_gr_steps :: IO GraphRewriteSystem -> IO TermGraph -> Int -> IO TermGraph
adv_gr_steps ugrs utg n | n <= 0 = error "At least one step is required."
                        | n == 1 = adv_gr_step ugrs utg 
                        | otherwise = do  
                                  grw <- ugrs
                                  tg  <- utg
                                  let result = Grewrite.informativeoutermostrewrite grw tg in 
                                      case result of
                                        Just (_, _, _, tg') -> putStrLn (show tg) >> information result >>  putStrLn " =>" >> (adv_gr_steps ugrs (return tg') (n-1)) --
                                        Nothing -> information result >> return tg
  

-- | adv_gr_nf ugrs utg: rewrites the given term utg wrt. graphrewritesystem ugrs straight to normal form 
adv_gr_nf :: IO GraphRewriteSystem -> IO TermGraph -> IO TermGraph
adv_gr_nf ugrs utg = do 
  grw <- ugrs
  tg  <- utg
  let result = Grewrite.outermostrewrite grw tg in 
    case result of
      Just tg' -> (adv_gr_nf ugrs (return tg')) 
      Nothing  -> return tg
  
-- | adv_sharing tg pos: shares tg maximally below pos
adv_sharing :: IO TermGraph -> Position -> IO TermGraph
adv_sharing utg pos =  do
  tg <- utg
  if (validpos tg pos)
     then return (foldtgatpos tg pos)
     else error "Invalid position."     

-- | adv_unsharing tg pos: unshares node in tg at position pos
adv_unsharing :: IO TermGraph -> Position -> IO TermGraph
adv_unsharing utg pos = do
  tg <- utg
  if (validpos tg pos) 
    then return (unfold tg pos)
    else error "Invalid position."


----------------------------------------------------------------
-- * advanced term functions
----------------------------------------------------------------

-- | adv_t_step utrw ut: one step rewrite of term ut wrt. trs utrs 
adv_t_step :: IO TermRewriteSystem -> IO UTerm -> IO UTerm
adv_t_step utrw ut = do
  t <- ut
  trw <- utrw 
  putStrLn $ show t
  putStrLn "->"
  case TermRewriting.outerRewrite trw t of
    []    -> putStrLn "Term is in normal form" >> return t 
    tj:_  -> return (TermRewriting.result tj)

-- | adv_t_steps utrw ut n: applies n rewritesteps to term ut wrt. trs utrw, halts if it reaches normal form before n steps
adv_t_steps :: IO TermRewriteSystem -> IO UTerm -> Int -> IO UTerm
adv_t_steps utrw ut n | n <= 0 = error "At least one step is required."
                      | n == 1 = adv_t_step utrw ut  
                      | otherwise = do
                                    t <- ut
                                    trw <- utrw  
                                    case TermRewriting.outerRewrite trw t of
                                      []    -> putStrLn "Term is in normal form" >> return t 
                                      tj:_  -> (putStrLn $ show t) >> (putStrLn "->") >> adv_t_steps utrw (return tj') (n-1) 
                                        where tj' = TermRewriting.result tj
                                              
-- | adv_t_nf utrw ut: rewrites term ut to normal form wrt. trs utrw
adv_t_nf :: IO TermRewriteSystem -> IO UTerm -> IO UTerm
adv_t_nf utrw ut = do 
  trw <- utrw
  t  <- ut
  let result = TermRewriting.outerRewrite trw t in 
    case result of
      tj:_ -> (adv_t_nf utrw (return $ TermRewriting.result tj)) 
      []  -> return t

----------------------------------------------------------------
-- * automatic term graph functions
----------------------------------------------------------------
 
-- | gr_step: performs a step with stored tg and grs and re-stores tg 
gr_step :: IO TermGraph
gr_step = do
  tg' <- adv_gr_step grs termgraph 
  set_termgraph tg'
  return tg'
  
  
-- | gr_steps n: performs n steps on stored termgraph and re-stores it globally 
--   if less than n steps are required to reach a normal form the computation  
--   halts before.  
gr_steps :: Int -> IO TermGraph 
gr_steps n = do
  tg' <- adv_gr_steps grs termgraph n
  set_termgraph tg'
  return tg'

-- | gr_nf: rewrites stored tg to normal form and re-stores it
--   i.e. all consecutive attempts to rewrite the stored tg without
--   changing the grs will be fruitless.
gr_nf :: IO TermGraph
gr_nf =  do
  tg' <- adv_gr_nf grs termgraph
  set_termgraph tg'
  return tg'
  
-- | sharing pos: shares the stored termgraph below position pos
sharing :: Position -> IO TermGraph
sharing pos = do
  tg' <- adv_sharing termgraph pos 
  return tg'

-- | unsharing pos: unshares the stored termgraph above position pos
unsharing :: Position -> IO TermGraph
unsharing pos = do
  tg' <- adv_unsharing termgraph pos 
  return tg'

----------------------------------------------------------------
-- * automatic term functions
----------------------------------------------------------------

-- | t_step: performs a step with stored term and trs
t_step :: IO UTerm
t_step = do
  t' <- adv_t_step trs term
  set_term t'
  return t'

-- | t_steps n: performs n steps on stored term and re-stores it globally 
--   if less than n steps are required to reach a normal form the computation  
--   halts before.  
t_steps :: Int -> IO UTerm 
t_steps n = do
  t' <- adv_t_steps trs term n
  set_term t'
  return t'
  
-- | t_nf: rewrites stored term to normalform and re-stores it.
t_nf :: IO UTerm
t_nf = do
  t' <- adv_t_nf trs term 
  set_term t'
  return t'

----------------------------------------------------------------
-- * storing operations
----------------------------------------------------------------

-- | set_termgraph tg: stores the given termgraph in STATE
set_termgraph :: TermGraph -> IO ()
set_termgraph tg = do
  STATE rw t _ <- getState
  writeIORef stateRef (STATE rw t (Just tg))
  
-- | set_term t: stores the given term in STATE
set_term :: UTerm -> IO ()
set_term t = do
  STATE rw _ tg <- getState
  writeIORef stateRef (STATE rw (Just t) tg)
  
  
----------------------------------------------------------------
-- * help
----------------------------------------------------------------

-- | help: helps
help :: IO () 
help = putDoc $
    b "Set the State of the Programm"  <$>    
    hang i (b "load_sys" <> text ":: FilePath -> IO ()" <$> 
               text "Parses a term rewrite system from the given file and stores it globally as graph rewrite system >grs and term rewrite system. >trs ") <$>
    hang i (b "load_term" <> text ":: String -> IO UTerm" <$> 
               text "Parses a term from the given String and stores it globally. >term ") <$>
    hang i (b "load_termgraph" <> text ":: String -> IO TermGraph" <$> 
               text "Parses a termgraph from the given (Term)-String and stores it globally. >termgraph ") <$>
    hang i (b "load_term" <> text ":: String -> IO (TermGraph, UTerm)" <$> 
               text "Parses a term and calculates the corresponding termgraph from the given String and stores both globally. >term and >termgraph ") <$>
    line <> 
    b "Inspect the Currently Stored Values"  <$>    
    hang i (b "term" <> text ":: IO UTerm" <$> 
               text "Returns the globally stored term. >load_term ") <$>
    hang i (b "termgraph" <> text ":: IO TermGraph" <$> 
               text "Returns the globally stored termgraph. >load_termgraph ") <$>
    hang i (b "trs" <> text ":: IO TermRewriteSystem" <$> 
               text "Returns the globally stored term rewrite system. >load_sys ") <$>
    hang i (b "grs" <> text ":: IO GraphRewriteSystem" <$> 
               text "Returns the globally stored graph rewrite system. >load_sys ") <$>
    line <> 
    b "Term Rewriting"  <$>    
    hang i (b "t_step" <> text ":: IO UTerm" <$> 
               text "Performs a single rewrite step on >term set by >load_term wrt. >trs set by >load_sys and stores the resulting term globally. ") <$>
    hang i (b "t_steps n" <> text ":: Int -> IO UTerm" <$> 
               text "Performs the given amount n of rewrite steps on >term set by >load_term wrt. >trs set by >load_sys and stores the resulting term globally. ") <$>
    hang i (b "t_nf" <> text ":: IO UTerm" <$> 
               text "Performs rewrite steps on >term set by >load_term wrt. >trs set by >load_sys until a normal form is reached and stores the resulting term globally. ") <$>
    line <> 
    b "Graph Rewriting"  <$>    
    hang i (b "gr_step" <> text ":: IO TermGraph" <$> 
               text "Performs a single rewrite step on >termgraph set by >load_termgraph wrt. >grs set by >load_sys and stores the resulting termgraph globally. ") <$>
    hang i (b "gr_steps n" <> text ":: Int -> IO TermGraph" <$> 
               text "Performs the given amount n of rewrite steps on >termgraph set by >load_termgraph wrt. >grs set by >load_sys and stores the resulting termgraph globally. ") <$>
    hang i (b "gr_nf" <> text ":: IO TermGraph" <$> 
            text "Performs rewrite steps on >termgraph set by >load_termgraph wrt. >grs set by >load_sys until a normal form is reached and stores the resulting termgraph globally.") <$>      
    hang i (b "sharing p" <> text ":: Position -> IO TermGraph" <$> 
               text "Maximally shares >termgraph set by >load_termgraph below given position p := [] | [int , position] ") <$>
    hang i (b "unsharing p" <> text ":: Position -> IO TermGraph" <$> 
            text "Unshares given position p := [] | [int , position] in >termgraph set by >load_termgraph. ") <$>
    line <> 
    b "Advanced (or Personalized) Term Rewriting"  <$>    
    hang i (b "adv_t_step trs t" <> text ":: IO TermRewriteSystem -> IO UTerm -> IO UTerm" <$> 
               text "Performs a single rewrite step on t  wrt. trs.") <$>
    hang i (b "adv_t_steps trs t n" <> text ":: IO TermRewriteSystem -> IO UTerm -> Int -> IO UTerm " <$> 
               text "Performs the given amount n of rewrite steps on t wrt. trs.") <$>
    hang i (b "t_nf trs t" <> text ":: IO TermRewriteSystem -> IO UTerm -> IO UTerm" <$> 
               text "Performs rewrite steps on t wrt. trs until a normal form is reached. ") <$>
    line <> 
    b "Advanced (or Personalized) Graph Rewriting"  <$>    
    hang i (b "adv_gr_step grs tg" <> text ":: IO GraphRewriteSystem -> IO TermGraph -> IO TermGraph" <$> 
               text "Performs a single rewrite step on tg wrt. grs. ") <$>
    hang i (b "adv_gr_steps grs tg n" <> text ":: IO GraphRewriteSystem -> IO TermGraph -> Int -> IO TermGraph" <$> 
               text "Performs the given amount n of rewrite steps on tg wrt. grs. ") <$>
    hang i (b "adv_gr_nf grs tg" <> text ":: IO GraphRewriteSystem -> IO TermGraph -> IO TermGraph" <$> 
            text "Performs rewrite steps on tg wrt. grs until a normal form is reached. ") <$>    
    hang i (b "sharing tg p" <> text ":: IO TermGraph -> Position -> IO TermGraph" <$> 
               text "Maximally shares tg below given position p := [] | [int , position] ") <$>
    hang i (b "unsharing tg p" <> text ":: TermGraph -> Position -> IO TermGraph" <$> 
            text "Unshares given position p := [] | [int , position] in tg. ")
    <> line 
     where  b str = bold $ text str 
            i = 4